﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using System.IO;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using System.Reflection;
using Microsoft.Xna.Framework.Content;

namespace EffectGeneratorTool
{
    /// <summary>
    /// Converts c# effect class code from a .fx effect file.
    /// </summary>
    public sealed class EffectGenerator
    {
        #region Properties

        private static ClassWriter ClassWriter { get; set; }

        private static readonly string[] ignoredSemantics = new string[] { "WorldViewProjection", "WorldViewProj", "World", "View", "Projection", "STANDARDSGLOBAL", "WorldInverseTranspose", "ViewInverse" };
        
        #endregion

        private EffectGenerator() {}

        public static string ConvertEffectToClass(Effect effect, string effectName, string namespaceName)
        {
            ClassWriter = new ClassWriter();

            // Write using and class header
            WriteUsingStatements(namespaceName);
            ClassWriter.WriteLine(String.Format("namespace {0}", namespaceName));
            ClassWriter.WriteLine("{");
            ClassWriter.TabRight();
            ClassWriter.Write(string.Format("class {0} : BaseEffect", effectName));
            ClassWriter.WriteLine();
            ClassWriter.WriteLine("{");
            ClassWriter.TabRight();

            WriteTechniqueEnumeration(effect.Techniques);

            WriteProperties(effect.Parameters);

            WriteConstructor(effect.Parameters, effectName);

            WriteOnApplyMethod(effect.Parameters, effect.Techniques);

            ClassWriter.TabLeft();
            ClassWriter.WriteLine("}"); // Class end

            ClassWriter.TabLeft();
            ClassWriter.WriteLine("}"); // Namespace end

            return ClassWriter.Code;
        }

        private static void WriteUsingStatements(string namespaceName)
        {
            ClassWriter.WriteLine("using System;");
            ClassWriter.WriteLine("using Microsoft.Xna.Framework;");
            ClassWriter.WriteLine("using Microsoft.Xna.Framework.Content;");
            ClassWriter.WriteLine("using Microsoft.Xna.Framework.Graphics;");
            ClassWriter.WriteLine(String.Format("using {0}.Interfaces;", namespaceName));
            ClassWriter.WriteLine();
        }

        private static void WriteConstructor(EffectParameterCollection effectParameterCollection, string effectName)
        {
            ClassWriter.WriteLine(String.Format("public {0}(GraphicsDevice graphicsDevice, byte[] effectCode) : base(graphicsDevice, effectCode)", effectName));
            ClassWriter.WriteLine("{");
            ClassWriter.TabRight();

            var materialName = effectName.Replace("Effect", "");
            ClassWriter.WriteLine(String.Format("Name = \"{0}\";", materialName));

            var filteredParameters = effectParameterCollection.Where(p => !ignoredSemantics.Contains(p.Semantic));

            foreach (var parameter in filteredParameters)
            {
                var hasValidValue = true;
                object value = String.Empty;

                #region Switch cases for parameter types: Convert HLSL value to C# value

                switch (parameter.ParameterType) {
                    case EffectParameterType.Bool:
                        value = parameter.GetValueBoolean();
                        break;
                    case EffectParameterType.Int32:
                        value = parameter.GetValueInt32();
                        break;
                    case EffectParameterType.Single:
                        if ((parameter.RowCount > 1) && (parameter.ColumnCount > 1)) // Ignore matrix values
                        {
                            hasValidValue = false;
                        }
                        else if (parameter.ColumnCount > 1)
                        {
                            var type = "Color";
                            switch (parameter.ColumnCount)
                            {
                                case 2:
                                    var vector2 = parameter.GetValueVector2();
                                    value = String.Format("new Vector2({0}f, {1}f)", FloatToString(vector2.X), FloatToString(vector2.Y));
                                    break;
                                case 3:
                                    if (!IsColor(parameter))
                                    {
                                        type = "Vector3";
                                    }
                                    var vector3 = parameter.GetValueVector3();
                                    value = String.Format("new {3}({0}f, {1}f, {2}f)", FloatToString(vector3.X), FloatToString(vector3.Y), FloatToString(vector3.Z), type);
                                    break;
                                case 4:
                                    if (!IsColor(parameter))
                                    {
                                        type = "Vector4";
                                    }
                                    var vector4 = parameter.GetValueVector4();
                                    value = String.Format("new {4}({0}f, {1}f, {2}f, {3}f)", FloatToString(vector4.X), FloatToString(vector4.Y), FloatToString(vector4.Z), FloatToString(vector4.W), type);
                                    break;
                                default:
                                    hasValidValue = false;
                                    break;
                            }
                        }
                        else
                        {
                            value = String.Format("{0}f", FloatToString(parameter.GetValueSingle()));
                        }
                        break;
                    case EffectParameterType.String:
                        value = parameter.GetValueString();
                        break;
                    case EffectParameterType.Texture:
                    case EffectParameterType.Texture2D:
                    case EffectParameterType.Texture1D:
                    case EffectParameterType.Texture3D:
                    case EffectParameterType.TextureCube:
                    default:
                        hasValidValue = false;
                        break;
                }

                #endregion

                if (hasValidValue) 
                {
                    ClassWriter.WriteLine(String.Format("{0} = {1};", GetUIName(parameter), value));
                }

            }

            ClassWriter.TabLeft();
            ClassWriter.WriteLine("}");
            ClassWriter.WriteLine();
        }

        private static void WriteProperties(EffectParameterCollection effectParameterCollection)
        {
            ClassWriter.WriteLine("#region Properties");
            ClassWriter.WriteLine();

            var filteredParameters = effectParameterCollection.Where(p => !ignoredSemantics.Contains(p.Semantic));

            foreach (EffectParameter parameter in filteredParameters)
            {
                string cSharpType = String.Empty;

                #region Switch cases for parameter types: Map HLSL types to C# types

                switch (parameter.ParameterType)
                {
                    case EffectParameterType.Bool:
                        cSharpType = "bool";
                        break;
                    case EffectParameterType.Int32:
                        cSharpType = "int";
                        break;
                    case EffectParameterType.Single:
                        if ((parameter.RowCount > 1) && (parameter.ColumnCount > 1))
                        {
                            cSharpType = "Matrix";
                        }
                        else if (parameter.ColumnCount > 1)
                        {
                            if (IsColor(parameter))
                            {
                                cSharpType = "Color";
                            }
                            else
                            {
                                cSharpType = "Vector" + parameter.ColumnCount.ToString();
                            }
                        }
                        else
                        {
                            cSharpType = "float";
                        }
                        break;
                    case EffectParameterType.String:
                        cSharpType = "string";
                        break;
                    case EffectParameterType.Texture:
                        if (parameter.Semantic.ToLower().Equals("ENVIRONMENT".ToLower()))
                        {
                            cSharpType = "TextureCube";
                        }
                        else
                        {
                            cSharpType = "Texture2D";
                        }
                        break;
                    case EffectParameterType.Texture1D:
                        cSharpType = "Texture1D";
                        break;
                    case EffectParameterType.Texture2D:
                        cSharpType = "Texture2D";
                        break;
                    case EffectParameterType.Texture3D:
                        cSharpType = "Texture3D";
                        break;
                    case EffectParameterType.TextureCube:
                        cSharpType = "TextureCube";
                        break;
                    default:
                        return;
                }

                #endregion

                // Is array?
                if (parameter.Elements.Count > 0)
                {
                    cSharpType = cSharpType + "[]";
                }

                var cSharpName = GetUIName(parameter);
                ClassWriter.WriteLine(string.Format("public {0} {1}", cSharpType, cSharpName) + " { get; set; }");
            }
            ClassWriter.WriteLine();

            ClassWriter.WriteLine("public EffectTechniques Technique { get; set; }");
            ClassWriter.WriteLine();
            ClassWriter.WriteLine("#endregion");
            ClassWriter.WriteLine();
        }

        private static void WriteOnApplyMethod(EffectParameterCollection effectParameterCollection, EffectTechniqueCollection effectTechniqueCollection)
        {
            ClassWriter.WriteLine("protected override void OnApply()");
            ClassWriter.WriteLine("{");
            ClassWriter.TabRight();

            #region Catch special semantics from NIVDIA Composer FX

            var filteredParameters = effectParameterCollection.Where(p => !(p.Semantic.ToLower().Equals("STANDARDSGLOBAL".ToLower())));
            
            foreach (var parameter in filteredParameters)
            {
                var semantik = parameter.Semantic.ToLower();

                var isWorldViewProjectionParameter = semantik.Equals("WorldViewProjection".ToLower()) || semantik.Equals("WorldViewProj".ToLower());
                if (isWorldViewProjectionParameter)
                {
                    ClassWriter.WriteLine(String.Format("Parameters[\"{0}\"].SetValue(World * View * Projection);", parameter.Name));
                }
                else if (semantik.Equals("WorldInverseTranspose".ToLower()))
                {
                    ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(Matrix.Transpose(Matrix.Invert(World)));");
                }
                else if (semantik.Equals("World".ToLower()))
                {
                    ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(World);");
                }
                else if (semantik.Equals("ViewInverse".ToLower()))
                {
                    ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(Matrix.Invert(View));");
                }
                else if (semantik.Equals("View".ToLower()))
                {
                    if (parameter.RowCount > 1)
                    {
                        ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(View);");
                    }
                    else
                    {
                        ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(new Vector4(View.M41, View.M42, View.M43, 1.0f));");
                    }
                }
                else if (semantik.Equals("Projection".ToLower()))
                {
                    ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(Projection);");
                }
                else if (semantik.Equals("Normal".ToLower()))
                {
                    ClassWriter.WriteLine(GetUIName(parameter) + ".Normalize();");
                    ClassWriter.WriteLine("Parameters[\"" + parameter.Name + "\"].SetValue(" + GetUIName(parameter) + ");");
                }
                else if (IsColor(parameter))
                {
                    ClassWriter.WriteLine(String.Format("Parameters[\"{0}\"].SetValue({1}.ToVector{2}());", parameter.Name, GetUIName(parameter), parameter.ColumnCount.ToString()));
                }
                else
                {
                    ClassWriter.WriteLine(String.Format("Parameters[\"{0}\"].SetValue({1});", parameter.Name, GetUIName(parameter)));
                }
            }
            ClassWriter.WriteLine();

            #endregion

            ClassWriter.WriteLine("switch (Technique)");
            ClassWriter.WriteLine("{");
            ClassWriter.TabRight();
            foreach (var technique in effectTechniqueCollection)
            {
                ClassWriter.WriteLine(string.Format("case EffectTechniques.{0}:", technique.Name));
                ClassWriter.TabRight();
                ClassWriter.WriteLine(string.Format("CurrentTechnique = Techniques[\"{0}\"];", technique.Name));
                ClassWriter.WriteLine("break;");
                ClassWriter.TabLeft();
            }
            if (effectParameterCollection.Count > 0)
            {
                ClassWriter.WriteLine("default:");
                ClassWriter.TabRight();
                ClassWriter.WriteLine(string.Format("CurrentTechnique = Techniques[\"{0}\"];", effectTechniqueCollection[0].Name));
                ClassWriter.WriteLine("break;");
                ClassWriter.TabLeft();
            }
            ClassWriter.TabLeft();
            ClassWriter.WriteLine("}");

            ClassWriter.TabLeft();
            ClassWriter.WriteLine("}");
        }

        private static void WriteTechniqueEnumeration(EffectTechniqueCollection effectTechniqueCollection)
        {
            ClassWriter.WriteLine("public enum EffectTechniques");
            ClassWriter.WriteLine("{");
            ClassWriter.TabRight();
            foreach (EffectTechnique technique in effectTechniqueCollection)
            {
                if (technique.Equals(effectTechniqueCollection.Last()))
                {
                    ClassWriter.WriteLine(string.Format("{0}", technique.Name));
                }
                else
                {
                    ClassWriter.WriteLine(string.Format("{0},", technique.Name));
                }
            }
            ClassWriter.TabLeft();
            ClassWriter.WriteLine("}");
            ClassWriter.WriteLine();
        }

        #region Helper

        private static bool IsColor(EffectParameter parameter)
        {
            return parameter.Semantic.ToLower().Equals("Color".ToLower()) || HasColorAnnotation(parameter);
        }

        private static string FloatToString(float value)
        {
            return value.ToString().Replace(",", ".");
        }

        private static bool HasColorAnnotation(EffectParameter parameter)
        {
            return parameter.Annotations["UIWidget"] != null && parameter.Annotations["UIWidget"].GetValueString().Equals("Color");
        }

        private static bool HasNameAnnotation(EffectParameter parameter)
        {
            return parameter.Annotations["UIName"] != null;
        }


        private static string GetUIName(EffectParameter parameter)
        {
            var cSharpName = parameter.Name;

            if (parameter.Annotations["UIName"] != null)
            {
                cSharpName = parameter.Annotations["UIName"].GetValueString().Replace(" ", "").Replace("-", "");
            }

            return cSharpName;
        }

        #endregion
    }
}
