#region Using Statements
using System;
using System.IO;
using System.Reflection;
using System.Windows.Forms;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;
using System.Collections.Generic;
#endregion

namespace EffectGeneratorTool
{
    /// <summary>
    /// Custom form provides the main user interface for the program.
    /// In this sample we used the designer to fill the entire form with a
    /// ModelViewerControl, except for the menu bar which provides the
    /// "File / Open..." option.
    /// </summary>
    public class Program
    {
        #region Directory and project names

        private static readonly string appProjectName = "XNATutorial";
        private static readonly string appEffectsDirectoryName = "Effects";
        private static readonly string contentProjectName = "XNATutorialContent";
        private static readonly string contentEffectsDirectoryName = "effects";

        private static readonly string appNamespaceName = String.Format("{0}.{1}", appProjectName, appEffectsDirectoryName);

        private static readonly string baseDirectory = Environment.CurrentDirectory.Substring(0, Environment.CurrentDirectory.IndexOf("EffectGenerator"));
        private static readonly string contentDirectory = String.Format("{0}{1}\\", baseDirectory, contentProjectName);
        private static readonly string effectsDirectory = String.Format("{0}{1}\\", contentDirectory, contentEffectsDirectoryName);
        private static readonly string outputDirectory = String.Format("{0}{1}\\{2}\\", baseDirectory, appProjectName, appEffectsDirectoryName);

        #endregion

        #region Properties

        private static ContentBuilder contentBuilder;
        private static ContentManager contentManager;

        private static List<Effect> effects;

        #endregion

        /// <summary>
        /// EffectGenerator entry point
        /// </summary>
        static void Main(string[] args)
        {
            #region Initialize

            contentBuilder = new ContentBuilder();

            var graphicsDeviceService = GraphicsDeviceService.AddRef(new Form().Handle, 1, 1);

            // Register the service, so components like ContentManager can find it.
            var services = new ServiceContainer();
            services.AddService<IGraphicsDeviceService>(graphicsDeviceService);

            contentManager = new ContentManager(services, contentBuilder.OutputDirectory);

            effects = new List<Effect>();

            #endregion

            var hardcodedNames = new List<string>(new string[] { "Environment" });

            #region Create Effect class files

            foreach (var effectName in EffectNamesFromEffectsDirectory())
            {
                GenerateEffect(effectName);
            }

            #endregion
        }

        private static List<string> EffectNamesFromEffectsDirectory()
        {
            var effectNames = new List<string>();
            foreach (var effectFilename in Directory.GetFiles(effectsDirectory))
            {
                var startIndex = effectFilename.LastIndexOf("\\") + 1;
                var length = effectFilename.IndexOf(".fx") - startIndex;
                var effectName = effectFilename.Substring(startIndex, length);
                effectNames.Add(effectName);
            }
            return effectNames;
        }

        /* Obsolete material enum generator method
        private static void CreateMaterialEnum()
        {
            var enumCode = MaterialGenerator.CreateEnum(EffectNamesFromEffectsDirectory(), appNamespaceName);

            try
            {
                var enumFile = String.Format("{0}Material.cs", outputDirectory);
                var streamWriter = new StreamWriter(enumFile);
                streamWriter.Write(enumCode);
                streamWriter.Flush();
                streamWriter.Close();
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }

        }
        */

        #region Helper methods

        /// <summary>
        /// Load a .fx file and convert a new c# effect-class file.
        /// </summary>
        /// <param name="effectFile">Name of the .fx effect-file to load without extension.</param>
        /// <param name="effectClassName">Name of the class to be generated.</param>
        /// <param name="interfaces">Optional: List of additional implemented interfaces of the generated class.</param>
        private static void GenerateEffect(string effectFilename)
        {
            var effectFullFilename = String.Format("{0}{1}.fx", effectsDirectory, effectFilename);
            var effect = LoadEffect(effectFullFilename);

            if (effect != null)
            {
                effects.Add(effect);

                var effectClassName = String.Format("{0}Effect", effectFilename);

                var outputFile = String.Format("{0}{1}.cs", outputDirectory, effectClassName);
                var classCode = EffectGenerator.ConvertEffectToClass(effect, effectClassName, appNamespaceName);

                try
                {
                    var streamWriter = new StreamWriter(outputFile);
                    streamWriter.Write(classCode);
                    streamWriter.Flush();
                    streamWriter.Close();
                }
                catch (Exception ex)
                {
                    Debug.WriteLine(ex.Message);
                }
            }
        }

        /// <summary>
        /// Loads an .fx effect-file.
        /// </summary>
        private static Effect LoadEffect(string fileName)
        {
            // Unload any existing effect.
            contentManager.Unload();

            // Tell the ContentBuilder what to build.
            contentBuilder.Clear();
            contentBuilder.Add(fileName, "Effect", null, "EffectProcessor");

            // Build this new model data.
            string buildError = contentBuilder.Build();

            Effect effect = null;

            if (string.IsNullOrEmpty(buildError))
            {
                // If the build succeeded, use the ContentManager to
                // load the temporary .xnb file that we just created.
                effect = contentManager.Load<Effect>("Effect");
            }
            else
            {
                // If the build failed, display an error message.
                MessageBox.Show(buildError, "Error");
            }

            return effect;
        }

        #endregion
    }
}
