using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using XNATutorial.Camera.Interfaces;

namespace XNATutorial.Camera
{
    /// <summary>
    /// Modified version from: http://hack1.codeplex.com/releases/view/52280#DownloadId=150242
    /// This is a camera object used to keep track of the position and rotation of a camera. Use of the camera
    /// will affect how models are shown in a scene.
    /// </summary>
    public class QuaternionCamera : ICamera
    {
        #region Properties

        /// <summary>
        /// Position of the camera in world-space.
        /// </summary>
        public Vector3 Position { get; protected set; }

        /// <summary>
        /// Rotation of the entity in object-space.
        /// </summary>
        public Quaternion Rotation { get; protected set; }

        /// <summary>
        /// The view matrix. Represents the viewing frustum. Needed to determine which objects in a 3D scene are displayed on the screen.
        /// </summary>
        public Matrix View { get; protected set; }

        /// <summary>
        /// The projection matrix. Needed to project a 3D scene on a 2D display plane.
        /// </summary>
        public Matrix Projection { get; protected set; }

        #endregion

        #region IMoveableEntity Interface

        /// <summary>
        /// Pans the entity along the X, Y, and Z directions.
        /// </summary>
        /// <param name="axis">The direction to pan the entity.</param>
        /// <param name="distance">The amount of X, Y, and Z distance to pan the entity. Units are assumed to be world units.</param>
        public void Translate(Vector3 axis, float distance)
        {
            var scaledDistance = axis * distance;
            Position += Vector3.Transform(scaledDistance, Matrix.CreateFromQuaternion(Rotation));
            UpdateView();
        }

        /// <summary>
        /// Rotates the entity along an arbitrary axis, relative to the position of the entity: object-space, not the world-space.
        /// </summary>
        /// <param name="axis">Ccontaining the axis along which to rotate the entity.</param>
        /// <param name="degrees">The number of degrees to rotate the entity by.</param>
        public void Rotate(Vector3 axis, float fDegrees)
        {
            // Transform rotation axis from world-space to object-space
            axis = Vector3.Transform(axis, Rotation);

            Rotation = Quaternion.CreateFromAxisAngle(axis, MathHelper.ToRadians(-fDegrees)) * Rotation;
            UpdateView();
        }

        /// <summary>
        /// Rotates the entity along an arbitrary axis, relative to the position of the entity: object-space, not the world-space.
        /// </summary>
        /// <param name="yaw">The number of degrees to rotate the entity arouhnd its y-axis.</param>
        /// <param name="pitch">The number of degrees to rotate the entity arouhnd its x-axis.</param>
        /// <param name="roll">The number of degrees to rotate the entity arouhnd its z-axis..</param>
        public void Rotate(float yaw, float pitch, float roll)
        {
            Rotation = Quaternion.CreateFromYawPitchRoll(MathHelper.ToRadians(yaw), MathHelper.ToRadians(pitch), MathHelper.ToRadians(roll)) * Rotation;
            UpdateView();
        }

        /// <summary>
        /// Creates view matrix from position and quaternion rotation.
        /// </summary>
        private void UpdateView()
        {
            View = Matrix.Invert(Matrix.CreateFromQuaternion(Rotation) * Matrix.CreateTranslation(Position));
        }

        #endregion

        #region Constructor & Initialization

        /// <summary>
        /// Creates a new instance of this object. 
        /// Camera is placed in front of the world-coordinate origin with no rotation, look down z-axis.
        /// </summary>
        /// <param name="newViewport">A Viewport object representing the view of the camera.</param>
        public QuaternionCamera(float fieldOfView, float aspectRatio, float nearClippingPlane, float farClippingPlane)
        {
            Position = new Vector3(0, 0, 1);
            Rotation = Quaternion.Identity;
            UpdateView();
            Perspective(fieldOfView, aspectRatio, nearClippingPlane, farClippingPlane);
        }

        /// <summary>
        /// Initializes the projection matrix.
        /// </summary>
        /// <param name="fieldOfView">Field of view.</param>
        /// <param name="aspectRatio">Aspect ratio of the window. See: GraphicsDevice.Viewport</param>
        /// <param name="nearClippingPlane">Near clipping plane of the view frustum.</param>
        /// <param name="farClippingPlane">Far clipping plane of the view frustum.</param>
        public void Perspective(float fieldOfView, float aspectRatio, float nearClippingPlane, float farClippingPlane)
        {
            Projection = Matrix.CreatePerspectiveFieldOfView(fieldOfView, aspectRatio, nearClippingPlane, farClippingPlane);
        }

        #endregion
    }
}
