﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;
using XNATutorial.Camera.Interfaces;
using XNATutorial.Effects;

namespace XNATutorial.Entity
{
    class ModelEntityComponent : DrawableGameComponent, IEntity
    {
        #region Properties

        /// <summary>
        /// Contains all visual mesh data. E.g. vertices, normals, vertex color, binormal. 
        /// The available data depdends on the loaded model.
        /// </summary>
        public Model Model { get; set; }

        /// <summary>
        /// Entity is drawn in relation to the current camera position.
        /// </summary>
        public ICamera Camera { get; set; }

        /// <summary>
        /// Position of the entity in world-space.
        /// </summary>
        public Vector3 Position { get; protected set; }

        /// <summary>
        /// Rotation of the entity in object-space.
        /// </summary>
        public Quaternion Rotation { get; protected set; }

        private Effect effect;

        /// <summary>
        /// The Effect determinates how the surface of the model is drawn to the screen.
        /// E.g. lighting, colors, textures, mappings, shadows, deformations.
        /// </summary>
        public Effect Effect
        {
            get
            {
                return effect;
            }
            set
            {
                Debug.Assert(Model != null, "Model Property is not initizalized.");
                
                effect = value;

                // Assign effect to model
                foreach (ModelMesh mesh in Model.Meshes)
                {
                    foreach (ModelMeshPart part in mesh.MeshParts)
                    {
                        part.Effect = value as Effect;
                    }
                }
            }
        }

        #endregion

        /// <summary>
        /// Creates a new instances of this object.
        /// </summary>
        /// <param name="game">Game that the game component should be attached to.</param>
        /// <param name="camera">Entity is drawn in relation to the current camera position.</param>
        public ModelEntityComponent(Game game, ICamera camera)
            : base(game)
        {
            Position = Vector3.Zero;
            Rotation = Quaternion.Identity;

            this.Camera = camera;
        }

        protected override void UnloadContent()
        {
            Effect.Dispose();

            base.UnloadContent();
        }

        #region IDrawable Interface

        /// <summary>
        /// Creates world matrix containing the position and the rotation of the entity in world-space.
        /// </summary>
        private Matrix World
        {
            get
            {
                // Howto extract forward- and up-vector from a quaternion:
                // http://www.gamedev.net/community/forums/topic.asp?topic_id=348394
                Vector3 forward = new Vector3(
                    -(2 * Rotation.X * Rotation.Z + 2 * Rotation.Y * Rotation.W),
                    -(2 * Rotation.Y * Rotation.Z - 2 * Rotation.X * Rotation.W),
                    -(1 - 2 * Rotation.X * Rotation.X - 2 * Rotation.Y * Rotation.Y)
                );
                Vector3 up = new Vector3(
                    2 * Rotation.X * Rotation.Y - 2 * Rotation.Z * Rotation.W,
                    1 - 2 * Rotation.X * Rotation.X - 2 * Rotation.Z * Rotation.Z,
                    2 * Rotation.Y * Rotation.Z + 2 * Rotation.X * Rotation.W
                );

                // Creates world matrix transformed from object-space to world-space.
                return Matrix.CreateWorld(Position, forward, up);
            }
        }

        public override void Draw(GameTime gameTime)
        {
            var effectMatrices = Effect as IEffectMatrices;
            Debug.Assert(effectMatrices != null, "Effect must implement IEffectMatrices interface");
            effectMatrices.World = World;
            effectMatrices.View = Camera.View;
            effectMatrices.Projection = Camera.Projection;

            foreach (var mesh in Model.Meshes)
            {
                mesh.Draw();
            }

            base.Draw(gameTime);
        }

        #endregion

        #region IMoveableEntity Interface

        /// <summary>
        /// Pans the entity along the X, Y, and Z directions.
        /// </summary>
        /// <param name="axis">The direction to pan the entity.</param>
        /// <param name="distance">The amount of X, Y, and Z distance to pan the entity. Units are assumed to be world units.</param>
        public void Translate(Vector3 axis, float distance)
        {
            Position += axis * distance;
        }

        /// <summary>
        /// Rotates the entity along an arbitrary axis, relative to the position of the entity: object-space, not the world-space.
        /// </summary>
        /// <param name="axis">Ccontaining the axis along which to rotate the entity.</param>
        /// <param name="fDegrees">The number of degrees to rotate the entity by.</param>
        public void Rotate(Vector3 axis, float degrees)
        {
            Rotation = Quaternion.CreateFromAxisAngle(axis, MathHelper.ToRadians(degrees)) * Rotation;
        }

        /// <summary>
        /// Rotates the entity along an arbitrary axis, relative to the position of the entity: object-space, not the world-space.
        /// </summary>
        /// <param name="yaw">The number of degrees to rotate the entity arouhnd its y-axis.</param>
        /// <param name="pitch">The number of degrees to rotate the entity arouhnd its x-axis.</param>
        /// <param name="roll">The number of degrees to rotate the entity arouhnd its z-axis..</param>
        public void Rotate(float yaw, float pitch, float roll)
        {
            Rotation = Quaternion.CreateFromYawPitchRoll(MathHelper.ToRadians(yaw), MathHelper.ToRadians(pitch), MathHelper.ToRadians(roll)) * Rotation;
        }

        #endregion
    }
}
