﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using XNATutorial.Camera.Interfaces;
using Microsoft.Xna.Framework.Graphics;
using XNATutorial.Effects;

namespace XNATutorial.Entity
{
    class SphereEntityComponent : ModelEntityComponent
    {
        private List<Effect> effects = new List<Effect>();
        private int effectsIndex = 0;

        private SpriteBatch spriteBatch;
        private SpriteFont font;

        public SphereEntityComponent(Game game, ICamera camera)
            : base(game, camera)
        {
        }

        protected override void LoadContent()
        {
            var content = Game.Content;

            // Create a new SpriteBatch, which can be used to draw font
            spriteBatch = new SpriteBatch(GraphicsDevice);
            font = content.Load<SpriteFont>("fonts/Default");

            // Load and set model
            Model = content.Load<Model>("models/Sphere");

            // Load effects
            var defaultDiffuseMap = Game.Content.Load<Texture2D>("textures/pak/metal/metal009_diffuse");
            var defaultNormalMap = Game.Content.Load<Texture2D>("textures/pak/metal/metal009_normal");

            var goochEffect = content.LoadDerivedEffect<GoochEffect>();
            effects.Add(goochEffect);

            var glossEffect = content.LoadDerivedEffect<GlossEffect>();
            glossEffect.DiffuseTexture = defaultDiffuseMap;
            glossEffect.GlossTexture = content.Load<Texture2D>("textures/pak/metal/metal009_specular");
            effects.Add(glossEffect);

            var lambertEffect = content.LoadDerivedEffect<LambertEffect>();
            lambertEffect.DiffuseTexture = defaultDiffuseMap;
            effects.Add(lambertEffect);

            var blinnEffect = content.LoadDerivedEffect<BlinnEffect>();
            blinnEffect.DiffuseTexture = defaultDiffuseMap;
            effects.Add(blinnEffect);

            var phongEffect = content.LoadDerivedEffect<PhongEffect>();
            phongEffect.DiffuseTexture = defaultDiffuseMap;
            phongEffect.SpecularPower = 30f;
            effects.Add(phongEffect);

            var basicEffect = new BasicEffect(Game.GraphicsDevice);
            basicEffect.Name = "BasicEffect (XNA)";
            basicEffect.EnableDefaultLighting();
            basicEffect.PreferPerPixelLighting = true;
            basicEffect.Texture = defaultDiffuseMap;
            basicEffect.TextureEnabled = true;
            basicEffect.SpecularPower = 30f;
            effects.Add(basicEffect);

            var environmentEffect = content.LoadDerivedEffect<EnvironmentEffect>();
            environmentEffect.Name = "Environment";
            environmentEffect.Texture = defaultDiffuseMap;
            environmentEffect.TextureEnabled = true;
            environmentEffect.EnvironmentMap = content.Load<TextureCube>("textures/environment/KitchenCube");
            effects.Add(environmentEffect);

            var environmentMapEffect = new EnvironmentMapEffect(Game.GraphicsDevice);
            environmentMapEffect.Name = "EnvironmentMapEffect (XNA)";
            environmentMapEffect.EnableDefaultLighting();
            environmentMapEffect.Texture = defaultDiffuseMap;
            environmentMapEffect.EnvironmentMap = content.Load<TextureCube>("textures/environment/KitchenCube");
            effects.Add(environmentMapEffect);

            var phongBumpReflectEffect = content.LoadDerivedEffect<PhongBumpReflectEffect>();
            phongBumpReflectEffect.DiffuseTexture = defaultDiffuseMap;
            phongBumpReflectEffect.NormalMapTexture = defaultNormalMap;
            phongBumpReflectEffect.SpecularPower = 30f;
            phongBumpReflectEffect.Environment = content.Load<TextureCube>("textures/environment/KitchenCube");
            effects.Add(phongBumpReflectEffect);

            // Set start effect
            SwitchNextEffect();

            base.LoadContent();
        }

        public void SwitchNextEffect()
        {
            Effect = effects[effectsIndex];
            effectsIndex++;
            if (effectsIndex >= effects.Count)
            {
                effectsIndex = 0;
            }
        }

        public override void Draw(GameTime gameTime)
        {
            // Draw sphere
            base.Draw(gameTime);

            // Draw debug text
            spriteBatch.Begin();
            var text = Effect.Name;
            var fontOrigin = font.MeasureString(text) / 2;
            spriteBatch.DrawString(font, text, fontOrigin + new Vector2(5f, 0), Color.LimeGreen, 0, fontOrigin, 1.0f, SpriteEffects.None, 0.5f);
            spriteBatch.End();

            // The SpriteBatch is changing some needed render states, so they are reset here.
            Game.GraphicsDevice.BlendState = BlendState.Opaque;
            Game.GraphicsDevice.DepthStencilState = DepthStencilState.Default;
        }
    }
}
