using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using XNATutorial.Entity;
using XNATutorial.Camera;
using XNATutorial.Camera.Interfaces;
using XNATutorial.Input;
using XNATutorial.Effects;
using XNATutorial.Effects.Interfaces;
using System.Diagnostics;

namespace XNATutorial
{
    /// <summary>
    /// This is the main type of the game
    /// </summary>
    public class GameApp : Game
    {
        #region Properties

        private TentacleEntityComponent TentacleEntityComponent { get; set; }
        private ModelEntityComponent BoxEntityComponent{ get; set; }

        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        public GameApp()
        {
            Window.Title = "XNA 4.0 Tutorial";

            Content.RootDirectory = "Content";

            #region Initialize graphics device

            var graphics = new GraphicsDeviceManager(this);

            graphics.PreferMultiSampling = true; // Turn on antialiasing 
            graphics.SynchronizeWithVerticalRetrace = true; // Turn on VSync
            graphics.GraphicsProfile = GraphicsProfile.HiDef; // Turn on best graphic settings

            graphics.IsFullScreen = false;

            if (graphics.IsFullScreen) // fullscreen
            {
                graphics.PreferredBackBufferWidth = 1920;
                graphics.PreferredBackBufferHeight = 1200;

                Window.AllowUserResizing = false;
            }
            else // window mode
            {
                graphics.PreferredBackBufferWidth = 1280;
                graphics.PreferredBackBufferHeight = 720;

                Window.AllowUserResizing = true;
            }

            graphics.ApplyChanges();

            #endregion
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content. Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // Initialize camera
            var camera = new QuaternionCamera(MathHelper.ToRadians(45.0f), GraphicsDevice.Viewport.AspectRatio, 1.0f, 1000.0f);
            camera.Translate(Vector3.Backward, 10); // Move camera backwards in front of the model. Camera looks directly down the z-axis to the model

            // Initialize entities
            TentacleEntityComponent = new TentacleEntityComponent(this, camera);
            BoxEntityComponent = new ModelEntityComponent(this, camera);
            BoxEntityComponent.Translate(Vector3.Right, 6.0f);

            // Initialize input components
            var cameraInputComponent = new CameraInputComponent(this, camera);
            var tentacleEntityInputComponent = new TentacleEntityInputComponent(this, TentacleEntityComponent);

            // Initialize game components
            Components.Add(cameraInputComponent);
            Components.Add(tentacleEntityInputComponent);
            Components.Add(TentacleEntityComponent);
            Components.Add(BoxEntityComponent);
            
            base.Initialize();
        }

        protected override void LoadContent()
        {
            BoxEntityComponent.Model = Content.Load<Model>("models/ChamferBox");
            var boxEntityEffect = new EnvironmentMapEffect(GraphicsDevice);
            boxEntityEffect.Texture = Content.Load<Texture2D>("textures/pak/metal/metal009_diffuse");
            boxEntityEffect.EnvironmentMap = Content.Load<TextureCube>("textures/environment/KitchenCube");
            boxEntityEffect.EnableDefaultLighting();
            boxEntityEffect.EnvironmentMapAmount = 0.5f;
            boxEntityEffect.EnvironmentMapSpecular = new Vector3(0.1f);
            BoxEntityComponent.Effect = boxEntityEffect;

            base.LoadContent();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed || Keyboard.GetState().IsKeyDown(Keys.Escape))
            {
                this.Exit();
            }

            // Continuously rotates the model
            var rotationStep = gameTime.ElapsedGameTime.Milliseconds * 0.01f;
            TentacleEntityComponent.Rotate(rotationStep, 0, rotationStep);
            BoxEntityComponent.Rotate(-rotationStep, rotationStep*2, rotationStep);

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Black);
           
            base.Draw(gameTime);
        }
    }
}
