using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using XNATutorial.Entity;

namespace XNATutorial.Input
{
    /// <summary>
    /// This is a game component that implements IUpdateable.
    /// </summary>
    class CameraInputComponent : GameComponent
    {
        /// <summary>
        /// Moveable entity object controlled by the user.
        /// </summary>
        public IMoveableEntity Entity { get; set; }

        private KeyboardState currentKeyboardState;
        private MouseState currentMouseState;

        /// <summary>
        /// Creates a new instance of this class.
        /// </summary>
        /// <param name="game">Game that the game component should be attached to.</param>
        /// <param name="camera">The camera controlled by the user.</param>
        public CameraInputComponent(GameApp game, IMoveableEntity camera) : base(game) 
        {
            Entity = camera;
        }

        /// <summary>
        /// Allows the game component to update itself. Catch user input and move the camera.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public override void Update(GameTime gameTime)
        {
            // Update input states
            currentKeyboardState = Keyboard.GetState();
            currentMouseState = Mouse.GetState();

            // Calculate viewport center point
            var viewportCenter = new Point(Game.GraphicsDevice.Viewport.Width / 2, Game.GraphicsDevice.Viewport.Height / 2);

            // Calculate mouse delta
            var mouseDelta = new Point(currentMouseState.X - viewportCenter.X, currentMouseState.Y - viewportCenter.Y);

            // Force mouse pointer back to center of the screen
            Mouse.SetPosition(viewportCenter.X, viewportCenter.Y);

            // Rotation
            var rotationMouseStep = 0.003f * gameTime.ElapsedGameTime.Milliseconds;

            if (mouseDelta.X != 0) 
            {
                Entity.Rotate(Vector3.UnitY, mouseDelta.X * rotationMouseStep);
            }

            if (mouseDelta.Y != 0)
            {
                Entity.Rotate(Vector3.UnitX, mouseDelta.Y * rotationMouseStep);
            }

            var rotationKeyboardStep = 0.02f * gameTime.ElapsedGameTime.Milliseconds;

            if (currentKeyboardState.IsKeyDown(Keys.Q))
            {
                Entity.Rotate(Vector3.UnitZ, rotationKeyboardStep);
            }

            if (currentKeyboardState.IsKeyDown(Keys.E))
            {
                Entity.Rotate(Vector3.UnitZ, -rotationKeyboardStep);
            }

            // Translation
            var translationStep = 0.005f * gameTime.ElapsedGameTime.Milliseconds;

            if (currentKeyboardState.IsKeyDown(Keys.W))
            {
                Entity.Translate(Vector3.Forward, translationStep);
            }

            if (currentKeyboardState.IsKeyDown(Keys.S))
            {
                Entity.Translate(Vector3.Backward, translationStep);
            }

            if (currentKeyboardState.IsKeyDown(Keys.A))
            {
                Entity.Translate(Vector3.Left, translationStep);
            }

            if (currentKeyboardState.IsKeyDown(Keys.D))
            {
                Entity.Translate(Vector3.Right, translationStep);
            }

            if (currentKeyboardState.IsKeyDown(Keys.R))
            {
                Entity.Translate(Vector3.Up, translationStep);
            }

            if (currentKeyboardState.IsKeyDown(Keys.F))
            {
                Entity.Translate(Vector3.Down, translationStep);
            }

            base.Update(gameTime);
        }
    }
}